@@ -33,6 +33,7 @@
 #include "gnome-vfs-mime-monitor.h"
 #include "gnome-vfs-monitor.h"
 #include "gnome-vfs-ops.h"
+#include "gnome-vfs-utils.h"
 #include "eggdesktopentries.h"
 #include "eggdirfuncs.h"
 
@@ -49,21 +50,19 @@ typedef struct {
 
 typedef struct {
 	GList *dirs;               /* mimeinfo.cache and defaults.list */
-	GList *defaults_list_dirs; /* defaults.list only dirs */
+	GHashTable *global_defaults_cache; /* global results of defaults.list lookup and validation */
         time_t last_stat_time;
 	guint should_ping_mime_monitor : 1;
 } GnomeVFSMimeInfoCache;
 
 extern void _gnome_vfs_mime_monitor_emit_data_changed (GnomeVFSMIMEMonitor *monitor); 
 
-static void free_mime_info_cache_map_list (GList *list);
 static void gnome_vfs_mime_info_cache_dir_init (GnomeVFSMimeInfoCacheDir *dir);
 static void gnome_vfs_mime_info_cache_dir_init_defaults_list (GnomeVFSMimeInfoCacheDir *dir);
 static GnomeVFSMimeInfoCacheDir *gnome_vfs_mime_info_cache_dir_new (const char *path);
 static void gnome_vfs_mime_info_cache_dir_free (GnomeVFSMimeInfoCacheDir *dir);
 static char **gnome_vfs_mime_info_cache_get_search_path (void);
-static char **gnome_vfs_mime_info_cache_get_defaults_search_path (void);
-;
+
 static gboolean gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (GnomeVFSMimeInfoCacheDir *dir,
 								      const char *desktop_entry);
 static void gnome_vfs_mime_info_cache_dir_add_desktop_entries (GnomeVFSMimeInfoCacheDir *dir,
@@ -76,11 +75,19 @@ static void gnome_vfs_mime_info_cache_free (GnomeVFSMimeInfoCache *cache);
 static GnomeVFSMimeInfoCache *mime_info_cache = NULL;
 G_LOCK_DEFINE_STATIC (mime_info_cache);
 
+
 static void
-free_mime_info_cache_map_list (GList *list)
+destroy_info_cache_value (gpointer key, GList *value, gpointer data)
 {
-	g_list_foreach (list, (GFunc) g_free, NULL);
-	g_list_free (list);
+	gnome_vfs_list_deep_free (value);
+}
+
+static void
+destroy_info_cache_map (GHashTable *info_cache_map)
+{
+	g_hash_table_foreach (info_cache_map, (GHFunc)destroy_info_cache_value,
+			      NULL);
+	g_hash_table_destroy (info_cache_map);
 }
 
 static gboolean
@@ -104,12 +111,28 @@ gnome_vfs_mime_info_cache_dir_out_of_date (GnomeVFSMimeInfoCacheDir *dir, const
 	return FALSE;
 }
 
+/* Call with lock held */
+static gboolean remove_all (gpointer  key,
+			    gpointer  value,
+			    gpointer  user_data)
+{
+	return TRUE;
+}
+
+
+static void
+gnome_vfs_mime_info_cache_blow_global_cache (void)
+{
+	g_hash_table_foreach_remove (mime_info_cache->global_defaults_cache,
+				     remove_all, NULL);
+}
+
 static void
 gnome_vfs_mime_info_cache_dir_init (GnomeVFSMimeInfoCacheDir *dir)
 {
 	EggDesktopEntries *entries;
 	GError *load_error;
-	gchar *filename, **desktop_file_ids, **mime_types;
+	gchar *filename, **mime_types;
 	int i;
 	static gchar *allowed_start_groups[] = { "MIME Cache", NULL };
 	struct stat buf;
@@ -125,12 +148,12 @@ gnome_vfs_mime_info_cache_dir_init (GnomeVFSMimeInfoCacheDir *dir)
 		return;
 
 	if (dir->mime_info_cache_map != NULL) {
-		g_hash_table_destroy (dir->mime_info_cache_map);
+		destroy_info_cache_map (dir->mime_info_cache_map);
 	}
 
 	dir->mime_info_cache_map = g_hash_table_new_full (g_str_hash, g_str_equal,
 							  (GDestroyNotify) g_free,
-							  (GDestroyNotify) free_mime_info_cache_map_list);
+							  NULL);
 
 	filename = g_build_filename (dir->path, "mimeinfo.cache", NULL);
 
@@ -162,6 +185,7 @@ gnome_vfs_mime_info_cache_dir_init (GnomeVFSMimeInfoCacheDir *dir)
 		goto error;
 
 	for (i = 0; mime_types[i] != NULL; i++) {
+		gchar **desktop_file_ids;
 		desktop_file_ids = egg_desktop_entries_get_string_list (entries,
 									"MIME Cache",
 									mime_types[i],
@@ -256,8 +280,6 @@ gnome_vfs_mime_info_cache_dir_init_defaults_list (GnomeVFSMimeInfoCacheDir *dir)
 		goto error;
 
 	for (i = 0; mime_types[i] != NULL; i++) {
-		GList *tmp;
-
 		desktop_file_id = egg_desktop_entries_get_string (entries,
 								  "Default Applications",
 								  mime_types[i],
@@ -267,35 +289,10 @@ gnome_vfs_mime_info_cache_dir_init_defaults_list (GnomeVFSMimeInfoCacheDir *dir)
 			load_error = NULL;
 			continue;
 		}
-
-		/* defaults.list in $XDG_CONFIG_DIRS/mime specify desktop
-		 * files relative to $XDG_DATA_DIRS/applications, so we
-		 * need to iterate through ever dir in $XDG_DATA_DIRS/applications
-		 * looking for a valid desktop file
-		 */
-		if (dir->mime_info_cache_map == NULL) {
-			tmp = mime_info_cache->dirs;
-			while (tmp != NULL) {
-				GnomeVFSMimeInfoCacheDir *app_dir;
-
-				app_dir = (GnomeVFSMimeInfoCacheDir *) tmp->data;
-				if (gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (app_dir,
-											  desktop_file_id)) {
-					g_hash_table_replace (dir->defaults_list_map,
-							      g_strdup (mime_types[i]),
-							      desktop_file_id);
-					break;
-				}
-				tmp = tmp->next;
-			}
-		} else {
-			if (gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (dir,
-										  desktop_file_id)) {
-				g_hash_table_replace (dir->defaults_list_map,
-						      g_strdup (mime_types[i]),
-						      desktop_file_id);
-			}
-		}
+		
+		g_hash_table_replace (dir->defaults_list_map,
+				      g_strdup (mime_types[i]),
+				      desktop_file_id);
 	}
 
 	g_strfreev (mime_types);
@@ -323,8 +320,23 @@ gnome_vfs_mime_info_cache_dir_changed (GnomeVFSMonitorHandle    *handle,
 		                       GnomeVFSMonitorEventType  event_type,
 	                               GnomeVFSMimeInfoCacheDir *dir)
 {
+	G_LOCK (mime_info_cache);
+	gnome_vfs_mime_info_cache_blow_global_cache ();
 	gnome_vfs_mime_info_cache_dir_init (dir);
+	G_UNLOCK (mime_info_cache);
+}
+
+static void
+gnome_vfs_mime_info_cache_dir_defaults_changed (GnomeVFSMonitorHandle    *handle,
+					 const gchar              *monitor_uri,
+					 const gchar              *info_uri,
+					 GnomeVFSMonitorEventType  event_type,
+					 GnomeVFSMimeInfoCacheDir *dir)
+{
+	G_LOCK (mime_info_cache);
+	gnome_vfs_mime_info_cache_blow_global_cache ();
 	gnome_vfs_mime_info_cache_dir_init_defaults_list (dir);
+	G_UNLOCK (mime_info_cache);
 }
 
 static GnomeVFSMimeInfoCacheDir *
@@ -352,7 +364,7 @@ gnome_vfs_mime_info_cache_dir_new (const char *path)
 			       filename,
 			       GNOME_VFS_MONITOR_FILE,
 			       (GnomeVFSMonitorCallback) 
-			       gnome_vfs_mime_info_cache_dir_changed,
+			       gnome_vfs_mime_info_cache_dir_defaults_changed,
 			       dir);
 	g_free (filename);
 
@@ -365,7 +377,7 @@ gnome_vfs_mime_info_cache_dir_free (GnomeVFSMimeInfoCacheDir *dir)
 	if (dir == NULL)
 		return;
 	if (dir->mime_info_cache_map != NULL) {
-		g_hash_table_destroy (dir->mime_info_cache_map);
+		destroy_info_cache_map (dir->mime_info_cache_map);
 		dir->mime_info_cache_map = NULL;
 
 	}
@@ -394,7 +406,7 @@ gnome_vfs_mime_info_cache_get_search_path (void)
 	char **args = NULL;
 	char **data_dirs;
 	char *user_data_dir;
-	int i, length;
+	int i, length, j;
 
 	data_dirs = egg_get_secondary_data_dirs ();
 
@@ -402,49 +414,22 @@ gnome_vfs_mime_info_cache_get_search_path (void)
 
 	args = g_new (char *, length + 2);
 
+	j = 0;
 	user_data_dir = egg_get_user_data_dir ();
-	args[length] = g_build_filename (user_data_dir, "applications", NULL);
+	args[j++] = g_build_filename (user_data_dir, "applications", NULL);
 	g_free (user_data_dir);
-	
-	i = length - 1;
-	while (i >= 0) {
-		args[length - i - 1] = g_build_filename (data_dirs[i],
-							 "applications", NULL);
-		i--;
+
+	for (i = 0; i < length; i++) {
+		args[j++] = g_build_filename (data_dirs[i],
+					      "applications", NULL);
 	}
-	args[length + 1] = NULL;
+	args[j++] = NULL;
 
 	g_strfreev (data_dirs);
 
 	return args;
 }
 
-static char **
-gnome_vfs_mime_info_cache_get_defaults_search_path (void)
-{
-	char **args = NULL;
-	char **config_dirs;
-	int i, length;
-
-	config_dirs = egg_get_secondary_configuration_dirs ();
-
-	for (length = 0; config_dirs[length] != NULL; length++);
-
-	args = g_new (char *, length + 1);
-
-	i = length - 1;
-	while (i >= 0) {
-		args[length - i - 1] = g_build_filename (config_dirs[i],
-							 "mime", NULL);
-		i--;
-	}
-	args[length] = NULL;
-
-	g_strfreev (config_dirs);
-
-	return args;
-}
-
 static gboolean
 gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (GnomeVFSMimeInfoCacheDir *dir,
                                                       const char *desktop_entry)
@@ -530,6 +515,24 @@ gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (GnomeVFSMimeInfoCacheDir *
 	return can_show_in;
 }
 
+static gboolean
+gnome_vfs_mime_info_desktop_entry_is_valid (const char *desktop_entry)
+{
+	GList *l;
+	
+	for (l = mime_info_cache->dirs; l != NULL; l = l->next) {
+		GnomeVFSMimeInfoCacheDir *app_dir;
+		
+		app_dir = (GnomeVFSMimeInfoCacheDir *) l->data;
+		if (gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (app_dir,
+									  desktop_entry)) {
+			return TRUE;
+		}
+	}
+	return FALSE;
+}
+
+
 static void
 gnome_vfs_mime_info_cache_dir_add_desktop_entries (GnomeVFSMimeInfoCacheDir *dir,
                                                    const char *mime_type,
@@ -542,9 +545,7 @@ gnome_vfs_mime_info_cache_dir_add_desktop_entries (GnomeVFSMimeInfoCacheDir *dir
 						mime_type);
 
 	for (i = 0; new_desktop_file_ids[i] != NULL; i++) {
-		if (!g_list_find (desktop_file_ids, new_desktop_file_ids[i]) &&
-		    gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (dir,
-									  new_desktop_file_ids[i]))
+		if (!g_list_find (desktop_file_ids, new_desktop_file_ids[i]))
 			desktop_file_ids = g_list_append (desktop_file_ids,
 							  g_strdup (new_desktop_file_ids[i]));
 	}
@@ -576,22 +577,6 @@ gnome_vfs_mime_info_cache_init_dir_lists (void)
 		}
 	}
 	g_strfreev (dirs);
-
-	dirs = gnome_vfs_mime_info_cache_get_defaults_search_path ();
-
-	for (i = 0; dirs[i] != NULL; i++) {
-		GnomeVFSMimeInfoCacheDir *dir;
-
-		dir = gnome_vfs_mime_info_cache_dir_new (dirs[i]);
-
-		if (dir != NULL) {
-			mime_info_cache->defaults_list_dirs =
-				g_list_append (mime_info_cache->defaults_list_dirs, dir);
-			gnome_vfs_mime_info_cache_dir_init_defaults_list (dir);
-		}
-	}
-
-	g_strfreev (dirs);
 }
 
 static void
@@ -604,23 +589,13 @@ gnome_vfs_mime_info_cache_update_dir_lists (void)
 
 		dir = (GnomeVFSMimeInfoCacheDir *) tmp->data;
 
-		if (dir->cache_monitor_handle == NULL) { 
+		if (dir->cache_monitor_handle == NULL) {
+			gnome_vfs_mime_info_cache_blow_global_cache ();
 			gnome_vfs_mime_info_cache_dir_init (dir);
 		}
 		
 		if (dir->defaults_monitor_handle == NULL) {
-			gnome_vfs_mime_info_cache_dir_init_defaults_list (dir);
-		}
-
-		tmp = tmp->next;
-	}
-
-	tmp = mime_info_cache->defaults_list_dirs;
-	while (tmp != NULL) {
-		GnomeVFSMimeInfoCacheDir *dir;
-
-		dir = (GnomeVFSMimeInfoCacheDir *) tmp->data;
-		if (dir->defaults_monitor_handle == NULL) { 
+			gnome_vfs_mime_info_cache_blow_global_cache ();
 			gnome_vfs_mime_info_cache_dir_init_defaults_list (dir);
 		}
 
@@ -668,6 +643,9 @@ gnome_vfs_mime_info_cache_new (void)
 
 	cache = g_new0 (GnomeVFSMimeInfoCache, 1);
 
+	cache->global_defaults_cache = g_hash_table_new_full (g_str_hash, g_str_equal,
+							      (GDestroyNotify) g_free,
+							      (GDestroyNotify) g_free);
 	return cache;
 }
 
@@ -681,10 +659,8 @@ gnome_vfs_mime_info_cache_free (GnomeVFSMimeInfoCache *cache)
 			(GFunc) gnome_vfs_mime_info_cache_dir_free,
 			NULL);
 	g_list_free (cache->dirs);
-	g_list_foreach (cache->defaults_list_dirs,
-			(GFunc) gnome_vfs_mime_info_cache_dir_free,
-			NULL);
-	g_list_free (cache->defaults_list_dirs);
+	g_hash_table_destroy (cache->global_defaults_cache);
+	g_free (cache);
 }
 
 void
@@ -777,9 +753,11 @@ gnome_vfs_mime_get_all_desktop_entries (const char *mime_type)
 			list = g_hash_table_lookup (dir->mime_info_cache_map, type);
 			
 			for (tmp = list; tmp != NULL; tmp = tmp->next) {
-				/* Add if not already in list */
+				/* Add if not already in list, and valid */
+				/* TODO: Should we cache here to avoid desktop file validation? */
 				if (!g_list_find_custom (desktop_entries, tmp->data,
-							 (GCompareFunc) strcmp)) {
+							 (GCompareFunc) strcmp) &&
+				    gnome_vfs_mime_info_desktop_entry_is_valid (tmp->data)) {
 					desktop_entries = g_list_prepend (desktop_entries,
 									  g_strdup (tmp->data));
 				}
@@ -799,48 +777,34 @@ gnome_vfs_mime_get_default_desktop_entry (const char *mime_type)
 {
 	gchar *desktop_entry;
 	GList *dir_list;
+	GnomeVFSMimeInfoCacheDir *dir;
 
 	gnome_vfs_mime_info_cache_init ();
 
 	G_LOCK (mime_info_cache);
-	dir_list = mime_info_cache->defaults_list_dirs;
-	desktop_entry = NULL;
-	while (dir_list != NULL) {
-		GnomeVFSMimeInfoCacheDir *dir;
 
-		dir = (GnomeVFSMimeInfoCacheDir *) dir_list->data;
-
-		desktop_entry = g_hash_table_lookup (dir->defaults_list_map,
-						     mime_type);
-
-		if (desktop_entry != NULL)
-			break;
-
-		dir_list = dir_list->next;
-	}
-	G_UNLOCK (mime_info_cache);
-
-	if (desktop_entry != NULL)
+	desktop_entry = g_hash_table_lookup (mime_info_cache->global_defaults_cache,
+					     mime_type);
+	if (desktop_entry) {
+		G_UNLOCK (mime_info_cache);
 		return g_strdup (desktop_entry);
+	}
 
-	G_LOCK (mime_info_cache);
-	dir_list = mime_info_cache->dirs;
 	desktop_entry = NULL;
-	while (dir_list != NULL) {
-		GnomeVFSMimeInfoCacheDir *dir;
-
-		dir = (GnomeVFSMimeInfoCacheDir *) dir_list->data;
-
+	for (dir_list = mime_info_cache->dirs; dir_list != NULL; dir_list = dir_list->next) {
+		dir = dir_list->data;
 		desktop_entry = g_hash_table_lookup (dir->defaults_list_map,
 						     mime_type);
-
-		if (desktop_entry != NULL)
-			break;
-
-		dir_list = dir_list->next;
+		if (desktop_entry != NULL &&
+		    gnome_vfs_mime_info_desktop_entry_is_valid (desktop_entry)) {
+			g_hash_table_insert (mime_info_cache->global_defaults_cache,
+					     g_strdup (mime_type), g_strdup (desktop_entry));
+			G_UNLOCK (mime_info_cache);
+			return g_strdup (desktop_entry);
+		}
 	}
 	G_UNLOCK (mime_info_cache);
 
-	return g_strdup (desktop_entry);
+	return NULL;
 }
 
