@@ -91,98 +91,111 @@
  * to avoid this...
  * Various ugly workarounds can be tried later.
  */
-dev_t
-_gnome_vfs_unix_mount_get_unix_device (GnomeVFSUnixMount *mount)
+GList *
+_gnome_vfs_unix_mount_get_unix_device (GList *mounts)
 {
-	struct stat statbuf;
-	dev_t unix_device;
-	pid_t pid;
-	int pipes[2];
-#ifdef HAVE_POLL
-	struct pollfd poll_fd;
-#else
-	struct timeval tv;
-	fd_set read_fds;
-#endif
-	int res;
-	int status;
+	GList *result = NULL;
 
-	if (pipe (pipes) == -1) {
-		return 0;
-	}
+	while (mounts) {
+		dev_t unix_device = 0;
+		pid_t pid;
+		int pipes[2];
+		int status;
 
-	pid = fork ();
-
-	if (pid == -1) {
-		close (pipes[0]);
-		close (pipes[1]);
-		return 0;
-	}
-
-	unix_device = 0;
-	
-	if (pid == 0) {
-		/* Child */
-		close (pipes[0]);
+		if (pipe (pipes) == -1)
+			goto error;
 
 		pid = fork ();
-
-		/* Fork an intermediate child that immediately exits so
-		 * we can waitpid it. This means the final process will get
-		 * owned by init and not go zombie
-		 */
-		if (pid == 0) {
-			/* Grandchild */
-			if (stat (mount->mount_path, &statbuf) == 0) {
-				write (pipes[1], (char *)&statbuf.st_dev, sizeof (dev_t));
-			}
-		} else {
+		if (pid == -1) {
+			close (pipes[0]);
 			close (pipes[1]);
+			goto error;
+		}
+
+		if (pid == 0) {
+			/* Child */
+			close (pipes[0]);
+
+			/* Fork an intermediate child that immediately exits
+			 * so we can waitpid it. This means the final process
+			 * will get owned by init and not go zombie.
+			 */
+			pid = fork ();
+
+			if (pid == 0) {
+				/* Grandchild */
+				struct stat statbuf;
+				while (mounts) {
+					GnomeVFSUnixMount *mount = mounts->data;
+					unix_device =
+						stat (mount->mount_path, &statbuf) == 0
+						? statbuf.st_dev
+						: 0;
+					write (pipes[1], (char *)&unix_device, sizeof (dev_t));
+					mounts = mounts->next;
+				}
+			}
+			close (pipes[1]);
+			_exit (0);
+			g_assert_not_reached ();
 		}
 
-		_exit (0);
-	} else {
 		/* Parent */
 		close (pipes[1]);
-		
-        wait_again:
+
+	retry_waitpid:
 		if (waitpid (pid, &status, 0) < 0) {
 			if (errno == EINTR)
-				goto wait_again;
+				goto retry_waitpid;
 			else if (errno == ECHILD)
 				; /* do nothing, child already reaped */
 			else
 				g_warning ("waitpid() should not fail in gnome_vfs_unix_mount_get_unix_device");
 		}
 
+		while (mounts) {
+			int res;
 
-		do {
+			do {
 #ifdef HAVE_POLL
-			poll_fd.fd = pipes[0];
-			poll_fd.events = POLLIN;
-			res = poll (&poll_fd, 1, STAT_TIMEOUT_SECONDS*1000);
+				struct pollfd poll_fd;
+				poll_fd.fd = pipes[0];
+				poll_fd.events = POLLIN;
+				res = poll (&poll_fd, 1, STAT_TIMEOUT_SECONDS * 1000);
 #else
-			tv.tv_sec = STAT_TIMEOUT_SECONDS;
-			tv.tv_usec = 0;
+				struct timeval tv;
+				fd_set read_fds;
+
+				tv.tv_sec = STAT_TIMEOUT_SECONDS;
+				tv.tv_usec = 0;
 			
-			FD_ZERO(&read_fds);
-			FD_SET(pipes[0], &read_fds);
+				FD_ZERO(&read_fds);
+				FD_SET(pipes[0], &read_fds);
 			
-			res = select (pipes[0] + 1,
-				      &read_fds, NULL, NULL, &tv);
+				res = select (pipes[0] + 1,
+					      &read_fds, NULL, NULL, &tv);
 #endif
-		} while (res == -1 && errno == EINTR);
-		
-		if (res > 0) {
-			if (read (pipes[0], (char *)&unix_device, sizeof (dev_t)) != sizeof (dev_t)) {
-				unix_device = 0; 
-			}
+			} while (res == -1 && errno == EINTR);
+
+			if (res <= 0 ||
+			    read (pipes[0], (char *)&unix_device, sizeof (dev_t)) != sizeof (dev_t))
+				break;
+
+			result = g_list_prepend (result, GUINT_TO_POINTER ((gulong)unix_device));
+			mounts = mounts->next;
 		}
 
 		close (pipes[0]);
+
+	error:
+		if (mounts) {
+			unix_device = 0;
+			result = g_list_prepend (result, GUINT_TO_POINTER ((gulong)unix_device));
+			mounts = mounts->next;
+		}
 	}
-	
-	return unix_device;
+
+	return g_list_reverse (result);
 }
 
 #ifndef HAVE_SETMNTENT
